/*
 * LM399_10V.c
 *
 * Created: 07.12.2018 20:40:08
 * Author : Andreas
 */ 
/*******************************************
Original code: adver (http://radiokot.ru/forum/viewtopic.php?p=3112844#p3112844)
Modified & ported to AVR Studio: Mickle T.
*******************************************
Compiler                : WinAVR
System Clock            : 5-20 MHz
Fuses					: XTAL, NO prescaler
*******************************************/
#define CODECHECK 0
#define POWERSAVE 1
#define HighSpeed 1

#define CLKTYPE 3
#if CLKTYPE == 3
#define F_CPU 2457600
#else
#if CLKTYPE == 2
#define F_CPU 4915200
#else 
  #if CLKTYPE == 1
#define F_CPU 8000000
  #else
#define F_CPU 20000000
  #endif
#endif
#endif

#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <avr/sleep.h>
//******************************************
#define BASIS 14            // 14 bits of fractional part (8+14bit total => 2.4uV resolution)
#define DIVISOR (1<<14)
#define Key_Up()   0 // !(PINB & ~(1<<PB1))		// Is "UP" pressed?
#define Key_Down() 0 // !(PINB & ~(1<<PB2))		// Is "DOWN" pressed?
//******************************************
#define UZ  6.86477
#define UVal (UZ*4194304UL/10)  // 2879293
unsigned long Value = UVal;            
// Value have a fixed point format [8.14]
// 256/Value = amplification factor of 7-10 V converter
// If LM399 output voltage = U_Zener, then 
// Value = (U_Zener/10.000 V)*256*(1<<BASIS) = (U_Zener/10.000 V)*4194304
//******************************************
// MASH internal variables
unsigned int ValInt;                        // Integer part of Value
unsigned int ValFrac;                       // Fractional part of Value

#if ((CODECHECK == 1) || (HighSpeed == 0))
unsigned int q[3]={1,0,0}; 	                // Latch outputs 16bit
char c[3][3]={{0,0,0}, {0,0,0}, {0,0,0}};   // Carry flip flops 1bit
unsigned int d[3];                          // Adder output / latch data input 16bit
#endif

#if ((CODECHECK == 1) || (HighSpeed == 1))
unsigned int q0 = 4;
unsigned int q1 = 0;
unsigned int q2 = 0;

unsigned char c1 = 0;
unsigned char c2 = 0;

unsigned int ValFrac4;
unsigned char ValIntM1;
#endif

#if CODECHECK == 1
unsigned long ErrCou = 0;
unsigned long Cycles = 0;
#endif
//******************************************
// Calibration factor (Value), saved in EEPROM
EEMEM unsigned long CFG_CAL_VALUE = UVal;
//******************************************
void Init(void) 
{
    DDRB |= (1 << PB0);                 // PWM output
    DDRB &= ~((1<<PB1) | (1<<PB2));     // UP & DOWN key pins config
    PORTB |= (1 << PB1) | (1 << PB2);   // Pullup resistors

#if ((POWERSAVE == 1) || (CLKTYPE == 3))
    /* power saving modes */
	ACSR   = 0x80; /* switch off analog comparator */
	ADMUX  = 0x20; /* VREF = VCC, MUX = ADC0 (temperature sensor), ADC left aligned */
	DIDR0  = 0x20; /* disable ADC0 digital input buffer */
#endif
#if CLKTYPE == 3
  ADCSRA = 0x04; /* ADC disabled clock = 2.4576 MHz / 2^4 */
#else
  #if CLKTYPE == 2
  ADCSRA = 0x05; /* ADC disabled clock = 4.9152 MHz / 2^5 */
  #else
    #if CLKTYPE == 1
  ADCSRA = 0x06; /* ADC disabled clock = 8 MHz / 2^6 */
    #else
  ADCSRA = 0x07; /* ADC disabled clock = 20 MHz / 2^7 */
    #endif
  #endif
#endif
//    OSCCAL = 0x97-10; /* 0x97 = nominal - 5% */
	
	/* variable init */
    ValInt = Value >> BASIS;
	ValIntM1 = ValInt - 1;
    ValFrac = Value % DIVISOR;
	ValFrac4 = ValFrac * 4;
    OCR0A = ValIntM1;          			// PWM compare match register initial value
    TCCR0A = (1<<COM0A1) | (1<<COM0A0) | (0<<COM0B1) | (0<<COM0B0) | (1<<WGM01) | (1<<WGM00);       
    // Set OC0A on Compare Match, BOTTOM, Normal mode, OC0B disconnected, Fast PWM Mode (WGM = 3) 
#if ((HighSpeed == 1) && (CODECHECK == 0))
    TCCR0B = (0<<WGM02) | (0<<CS02) | (0<<CS01) | (1<<CS00);    // Prescaler CK/1
#else	
    TCCR0B = (0<<WGM02) | (0<<CS02) | (1<<CS01) | (0<<CS00);    // Prescaler CK/8 
#endif
    TIMSK = (1 << TOIE0);  				// Enable overflow interrupt
    sei();
}

static inline unsigned char rotate_left_u8 (unsigned char x);
static inline unsigned char rotate_left_u8 (unsigned char x)
{
	asm ("rol %0" : "+r" (x));
	return x;
}
	
//******************************************
ISR (TIM0_OVF_vect)
{
#if ((HighSpeed == 1) || (CODECHECK == 1))
    register unsigned char c0;
	
	q0 = q0 + ValFrac4;
	c0 = ValIntM1 + rotate_left_u8(0u);
	
	q1 = q1 + q0;
	c1 = rotate_left_u8(c1);
    if (c1 & 1) c0++;
	if (c1 & 2) c0--;
	
	q2 = q2 + q1;
    c2 = rotate_left_u8(c2);
	if (c2 & 1) c0 ++;
	if (c2 & 2) c0 -= 2;
	if (c2 & 4) c0 ++;

	OCR0A = c0;
	
#endif

#if ((HighSpeed == 0) || (CODECHECK == 1))	
    // Carry flip-flops
    c[1][1] = c[1][0];  c[2][2] = c[2][1]; c[2][1] = c[2][0];
    // Adders
    d[0] = (q[0]+ValFrac) % DIVISOR;
    d[1] = (q[1]+d[0]) % DIVISOR;
    d[2] = (q[2]+d[1]) % DIVISOR;
    // Carries
    c[0][0] = (q[0]+ValFrac) >> BASIS;
    c[1][0] = (q[1]+d[0]) >> BASIS;
    c[2][0] = (q[2]+d[1]) >> BASIS;
    // Latch
    for (int i=0; i<3; i++) q[i] = d[i];
    OCR1A = ValInt - 1 
            + c[0][0]
            + c[1][0] -   c[1][1]
            + c[2][0] - 2*c[2][1] +   c[2][2];
#endif	

#if (CODECHECK == 1)
    if (OCR1A != OCR0A) ErrCou ++;
	Cycles ++;
#endif	
}
//******************************************
// PWM configuration update
//****************************************** 
void Update(void) 
{
    cli();
    ValInt = Value >> BASIS;
	ValIntM1 = ValInt-1;
    ValFrac = Value % DIVISOR;
	ValFrac4 = ValFrac * 4;
    OCR0A = ValIntM1;          
    sei();
}
//******************************************
// Loading the calibration factor (Value) from EEPROM
//******************************************
void ConfigLoad (void)
{
// If the both keys was pressed on the startup,
// only the failsafe initial value is loaded !!!
    if (!Key_Up() && !Key_Down()) {
      eeprom_read_block(&Value,&CFG_CAL_VALUE,4);
//      _delay_ms(3000); 
    };
}
//******************************************
// Update calibration factor in EEPROM
//******************************************
void ConfigUpdate (void)
{
    eeprom_update_block(&Value,&CFG_CAL_VALUE,4);
    _delay_ms(3000);
}
//******************************************
int main( void ) 
{
#if (CODECHECK == 0)
    /* EEPROM simulation may not work correctly in debugger */
    ConfigLoad();					// If UP & DOWN keys was pressed at startup
#endif								// then load default config 
    Init();
#if POWERSAVE == 1	
    set_sleep_mode(SLEEP_MODE_IDLE);
#endif
    while(1) {
      if (Key_Up() && Key_Down()) 	// If UP & DOWN keys was pressed
	    ConfigUpdate(); 			// then save config to EEPROM
      if (Key_Up() && !Key_Down()) {
        Value ++;
        Update(); };
      if (!Key_Up() && Key_Down()) {
        Value --;
        Update(); };
//     _delay_ms(300);
#if POWERSAVE == 1
     sleep_mode();
#endif 
    };
}
